COSMO = "Planck18"
IMF_TYPE = 1
IMF_UPPER_LIMIT = 120.0
IMF_LOWER_LIMIT = 0.08
IMF1 = 1.3
IMF2 = 2.3
IMF3 = 2.3
VDMC = 0.08
MDAVE = 0.5
IGM_TYPE = 0
LOG_XI_ION = 25.39
EPSILON = 0.3
DUST_INDEX = 0.0
DUST_INDEX_BC = -0.7
T_ESC = 0.01 
DUST_ETA = 1.0
SCALE_DUST_REDSHIFT = "Vogelsberger20"
DUST_LAW = 0
BUMP_AMP = 0.85
BUMP_DWAVE = 0.035
SALIM_A0 = -4.30
SALIM_A1 = 2.71
SALIM_A2 = -0.191
SALIM_A3 = 0.0121
SALIM_RV = 3.15
SALIM_B = 1.57